import bpy
import os
from . import fn

class RT_PG_selectable_prop(bpy.types.PropertyGroup):
    node_name: bpy.props.StringProperty(name="Node Name")
    node_label: bpy.props.StringProperty(name="Node Label")
    name: bpy.props.StringProperty(name="Name or Path")
    socket_name: bpy.props.StringProperty(name="Source socket Name") # Source socket name as reference
    select: bpy.props.BoolProperty(name="Selected", default=True)
    is_linked: bpy.props.BoolProperty(name="Linked", default=False)
    # is_valid: bpy.props.BoolProperty(name="Valid", default=True)
    
    ## extra output naming options
    name_from_node: bpy.props.StringProperty(name="Name From Node")
    name_from_node_and_socket: bpy.props.StringProperty(name="Name From Node And Socket")

class RT_OT_connect_selected_to_file_out(bpy.types.Operator):
    bl_idname = "rt.connect_selected_to_file_out"
    bl_label = "Connect Selected To File Output"
    bl_description = "Connect Selected Nodes to a new fileoutput node\
        \nIf a fileoutput node is selected, socket are added to it"
    bl_options = {"REGISTER", "UNDO"}

    socket_collection : bpy.props.CollectionProperty(type=RT_PG_selectable_prop)

    show_custom_settings : bpy.props.BoolProperty(
        name='Settings',
        default=False)

    ## enum choice for naming: socket_name, node_name, node_and_socket_name, 
    name_type : bpy.props.EnumProperty(
        name='Output Name From',
        description='Choose the output name\
            \nNode name use Label (Use node name when there is no Label)',
        default='node_and_socket_name',
        items=(
                ('node_and_socket_name', 'Node_Socket Name', 'Use the node name prefix and socket name', 0),
                ('socket_name', 'Socket Name', 'Use the socket name as output name', 1),
                ('node_name', 'Node Name', 'Use the node name as output name', 2),
            )
        )

    # prefix_with_node_name : bpy.props.BoolProperty(
    #     name='Prefix With Node Name',
    #     description='Add the node name as prefix to the output name',
    #     default=False)

    base_path : bpy.props.StringProperty(
        name='Custom base path',
        default='',
        description='Set the base path of created file_output (not if already exists)')
    
    file_format : bpy.props.EnumProperty(
        name='Output Format',
        default='NONE',
        items=(
                ('NONE', 'Default', 'Use default settings'),
                ('OPEN_EXR_MULTILAYER', 'OpenEXR MultiLayer', 'Output image in multilayer OpenEXR format'),
                ('OPEN_EXR', 'OpenEXR', 'Output image in OpenEXR format'),
            )
        )

    exr_codec : bpy.props.EnumProperty(
        name='Codec',
        default='PIZ',
        description='Codec settings for OpenEXR',
        items=(
            ('PIZ', 'PIZ (lossless)', ''),
            ('ZIP', 'ZIP (lossless)', ''),
            ('RLE', 'RLE (lossless)', ''),
            ('ZIPS', 'ZIPS (lossless)', ''),
            ('PXR24', 'Pxr24 (lossy)', ''),
            ('B44', 'B44 (lossy)', ''),
            ('B44A', 'B44A (lossy)', ''),
            ('DWAA', 'DWAA (lossy)', ''),
            ('DWAB', 'DWAB (lossy)', ''),
            ),
        )
    
    color_depth : bpy.props.EnumProperty(
        name='Color Depth',
        default='16',
        description='Bit depth per channel',
        items=(
                # ('8', '8', '8-bit color channels'),
                # ('10', '10', '10-bit color channels'),
                # ('12', '12', '12-bit color channels'),
                ('16', '16 (Half)', '16-bit color channels'),
                ('32', '32 (Full)', '32-bit color channels'),
            ),
        )

    def invoke(self, context, event):
        self.socket_collection.clear()
        if event.ctrl:
            # Direct connect, do not use any options
            self.base_path = ''
            return self.execute(context)

        selected = [n for n in context.scene.node_tree.nodes if n.select and n.type != 'OUTPUT_FILE']
        if not selected:
            self.report({'ERROR'}, 'No render layer nodes selected')
            return {'CANCELLED'}
        for n in selected:
            for o in n.outputs:
                if o.is_unavailable:
                    continue
                item = self.socket_collection.add()
                item.node_name = n.name
                item.node_label = n.label.strip()
                item.socket_name = o.name
                
                ## Set editable names
                item.name = o.name

                ## Store other naming options (cleaned at exec with bpy.path.clean_name)
                node_name = n.label.strip() if n.label.strip() else n.name
                
                ## Change node_name for render layers: scene_viewlayer_name
                if n.type == 'R_LAYERS' and node_name != n.label: # skip if a label is set
                    node_name = f'{n.layer}'
                    # node_name = f'{n.scene.name}_{n.layer}'

                item.name_from_node = node_name

                if len(n.outputs) == 1:
                    ## Only one output, just pick node name, no need to add socket name
                    item.name_from_node_and_socket = node_name
                else:
                    item.name_from_node_and_socket = f'{node_name}_{o.name}'

                ## TODO: rename item.name according to template pairs in preferences (to add later)
                if o.is_linked:
                    item.is_linked = True
                    item.select = False

        return context.window_manager.invoke_props_dialog(self, width=500)

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        expand_icon = 'DISCLOSURE_TRI_DOWN' if self.show_custom_settings else 'DISCLOSURE_TRI_RIGHT'
        box.prop(self, 'show_custom_settings', emboss=False, icon=expand_icon)
        ## Settings
        if self.show_custom_settings:
            box.use_property_split = True
            box.row().prop(self, 'name_type', expand=False)
            box.prop(self, 'base_path')
            col = box.column()
            col.prop(self, 'file_format')
            col.prop(self, 'exr_codec')
            col.row().prop(self, 'color_depth', expand=True)

        ## Node Sockets
        layout.use_property_split = False
        col = layout.column()
        current_node_name = ''
        for item in self.socket_collection:
            if item.node_name != current_node_name:
                current_node_name = item.node_name
                col.separator()
                ## Display node label + name or name only
                if item.node_label:
                    display_node_name = f'{item.node_label} ({item.node_name})'
                else:
                    display_node_name = item.node_name
                    ## A bit dirty: For render layer node, show render layer-node name.
                    if display_node_name.startswith('Render Layer'):
                        rln = context.scene.node_tree.nodes.get(item.node_name)
                        display_node_name = f'{rln.scene.name}/{rln.layer} ({display_node_name})'
                col.label(text=display_node_name, icon='NODE_SEL')

            row = col.row()
            if item.is_linked:
                row.label(text='', icon='LINKED') # NODETREE
            else:
                row.label(text='', icon='BLANK1')
            row.prop(item, 'select', text='')

            display_name = item.socket_name
            if 'crypto' in display_name.lower():
                display_name = f'{display_name} -> 32bit output node'
            row.label(text=display_name)
            row.label(text='', icon='RIGHTARROW')

            if self.name_type == 'socket_name':
                row.prop(item, 'name', text='')
            elif self.name_type == 'node_name':
                row.prop(item, 'name_from_node', text='')
            elif self.name_type == 'node_and_socket_name':
                row.prop(item, 'name_from_node_and_socket', text='')

    def execute(self, context):
        
        # Build exclude dict from selection
        excludes = {}
        remap_names = {}
        ## Old system
        # if len(self.socket_collection):
        #     for item in self.socket_collection:
        #         if not item.select:
        #             # All deselected goes to exclude with {node_name: [socket1, ...]}
        #             excludes.setdefault(item.node_name, []).append(item.name)
        #         elif item.socket_name != item.name:
        #             remap_names[item.socket_name] = item.name
        
        if len(self.socket_collection):
            for item in self.socket_collection:
                final_name = item.name
                ## change name if other options were used
                if self.name_type == 'node_name':
                    final_name = item.name_from_node
                elif self.name_type == 'node_and_socket_name':
                    final_name = item.name_from_node_and_socket
                
                if not item.select:
                    # All deselected goes to exclude with {node_name: [socket1, ...]}
                    excludes.setdefault(item.node_name, []).append(item.socket_name)
                elif item.socket_name != final_name:
                    remap_names.setdefault(item.node_name, {})[item.socket_name] = final_name
                    # remap_names[item.socket_name] = final_name

        ## Handle default file format
        file_ext = self.file_format
        if self.file_format == 'NONE':
            env_file_format = os.environ.get('FILE_FORMAT')
            file_ext = env_file_format if env_file_format else 'OPEN_EXR_MULTILAYER'

        file_format = {
            'file_format' : file_ext,
            'exr_codec' : self.exr_codec,
            'color_depth' : self.color_depth,
        }
        
        scn = context.scene
        nodes = scn.node_tree.nodes
        selected = [n for n in nodes if n.select]
        outfile = next((n for n in selected if n.type == 'OUTPUT_FILE'), None)
        # Exclude output file from 
        selected = [n for n in selected if n.type != 'OUTPUT_FILE']        
        
        # fn.connect_to_file_output(selected, outfile)
        for n in selected:
            fn.connect_to_file_output(n, outfile, base_path=self.base_path, excludes=excludes, remap_names=remap_names, file_format=file_format)
        return {"FINISHED"}

classes=(
RT_PG_selectable_prop,
RT_OT_connect_selected_to_file_out,
)

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

def unregister():
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)